"""Predict reference period."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import logging

# externals
import torch
import numpy as np
import xarray as xr
from torch.utils.data import DataLoader

# locals
from climax.core.dataset import EoDataset, NetCDFDataset

# module level logger
LOGGER = logging.getLogger(__name__)


def predict_ERA5(net, ERA5_ds, predictand, loss, batch_size=16, **kwargs):
    # create PyTorch compliant dataset and dataloader instances for model
    # inference
    ds = NetCDFDataset(ERA5_ds, **kwargs)
    dl = DataLoader(ds, batch_size=batch_size, shuffle=False, drop_last=False)

    # device to compute on: use gpu if available
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
    LOGGER.info('Predicting on device: {}'.format(device))

    # check if multiple gpus are available
    if torch.cuda.device_count() > 1:
        LOGGER.info('Using {} available GPUs.'.format(
            torch.cuda.device_count()))
        net = torch.nn.DataParallel(net)

    # copy network to target device and set to evaluation mode
    net = net.to(device)
    LOGGER.info('Setting model to evaluation mode ...')
    net.eval()

    # iterate over the reference period
    target = None
    for batch, (inputs, _) in enumerate(dl):

        # send the data to the gpu if available
        inputs = inputs.to(device)

        # calculate network outputs
        with torch.no_grad():
            outputs = net(inputs).cpu().numpy().squeeze()

        # stack predictions
        target = (np.concatenate((target, outputs), axis=0) if batch > 0 else
                  outputs)
        LOGGER.info('Mini-batch: {:d}/{:d}'.format(batch + 1, len(dl)))

    # check how many output fields are modelled
    if net.classifier.out_channels == 1:
        # single output field
        ds = {predictand: target}
    else:
        # joint optimization of min and max temperature
        if predictand == 'tas' and net.classifier.out_channels == 2:
            # in case of tas, the network predicts both tasmax and tasmin
            ds = {'tasmax': target[:, 0, ...].squeeze(),
                  'tasmin': target[:, 1, ...].squeeze()}

        # probabilistic precipitation modelling
        elif predictand == 'pr' and net.classifier.out_channels == 3:

            # probability of precipitation
            prob = torch.sigmoid(torch.as_tensor(target[:, 0, ...].squeeze(),
                                                 dtype=torch.float32)).numpy()

            # check which loss function is used

            # Bernoulli-Gamma
            if loss.__class__.__name__ == 'BernoulliGammaLoss':
                # precipitation amount: expected value of Bernoulli-Gamma
                # distribution
                # pr = p * shape * scale
                pr = (prob * np.exp(target[:, 1, ...].squeeze()) *
                      np.exp(target[:, 2, ...].squeeze()))

            # Bernoulli-GenPareto
            if loss.__class__.__name__ == 'BernoulliGenParetoLoss':
                # precipitation amount: expected value of Bernoulli-GenPareto
                # distribution
                # pr =
                # pr = (prob * np.exp(target[:, 1, ...].squeeze()) *
                #       np.exp(target[:, 2, ...].squeeze()))
                pass

            ds = {'prob': prob, 'precipitation': pr}

    # add coordinates to arrays
    ds = {k: EoDataset.add_coordinates(v) for k, v in ds.items()}

    # create xarray dataset: dtype=Float32
    ds = xr.Dataset(data_vars=ds,
                    coords={'time': ERA5_ds.time, 'y': ERA5_ds.y,
                            'x': ERA5_ds.x})

    return ds
